from abc import ABC, abstractmethod
from enum import Enum
from typing import Set, Text, Union

from ithildin.contract.loader import FileLoader, BinaryLoader, SolidityLoader, JsonRpcLoader


class LoaderFactoryType(Enum):
    BINARY = 1
    SOLIDITY = 2
    JSON_RPC = 3


class ContractLoaderFactory(ABC):

    def __init__(self, **options) -> None:
        missing_options = self._required_options - options.keys()
        if len(missing_options) > 0:
            raise AttributeError('Missing option(s): %s' % str(missing_options))
        self._options = options

    @property
    @abstractmethod
    def _required_options(self) -> Set[Text]:
        pass

    @abstractmethod
    def create(self) -> Union[FileLoader, JsonRpcLoader]:
        pass


class BinaryLoaderFactory(ContractLoaderFactory):

    def create(self) -> FileLoader:
        return BinaryLoader(self._options.get('path'))

    @property
    def _required_options(self) -> Set[Text]:
        return {'path'}


class SolidityLoaderFactory(ContractLoaderFactory):

    def create(self) -> FileLoader:
        return SolidityLoader(self._options.get('path'), solc=self._options.get('solc'))

    @property
    def _required_options(self) -> Set[Text]:
        return {'path', 'solc'}


class JsonRpcLoaderFactory(ContractLoaderFactory):

    def create(self) -> JsonRpcLoader:
        return JsonRpcLoader(self._options.get('address'), self._options.get('rpc'))

    @property
    def _required_options(self) -> Set[Text]:
        return {'address', 'rpc'}


def get_factory(loader_type: LoaderFactoryType, **options) -> ContractLoaderFactory:
    switcher = {
        LoaderFactoryType.BINARY:   BinaryLoaderFactory,
        LoaderFactoryType.SOLIDITY: SolidityLoaderFactory,
        LoaderFactoryType.JSON_RPC: JsonRpcLoaderFactory
    }
    if loader_type not in switcher:
        raise NotImplementedError('This factory has not been implemented yet')
    return switcher.get(loader_type)(**options)



import logging

from typing import Optional
from mythril.laser.ethereum.state.global_state import GlobalState

from ithildin.analysis.base import AnalysisStrategy
from ithildin.report.analysis import Result

log = logging.getLogger(__name__)


class CallData:
    """ Annotation for input data. """
    pass


class HashedInput:
    """ Annotation to be used on SHA3 elements. """
    pass


class HashedStorage:
    """ Annotation to be used on SLOAD elements, where the lookup key has been hashed. """
    pass


class HashLock(AnalysisStrategy):

    pattern_name = 'HASH_LOCK'
    report_title = 'Hash Lock (Off-Chain Secret Enabled Authentication)'
    report_description = ('This pattern aims at providing authentication when the account is not known in advance. '
                          'A secret is fixed off-chain and hashed, and the hash gets submitted to the contract. '
                          'Later, the holder of the secret can submit it to the contract, the contract checks '
                          'the secret\'s hash against the stored one, and if they match the protected logic gets executed.')

    pre_hooks = ['JUMPI', 'SLOAD']
    post_hooks = ['SHA3', 'SLOAD', 'CALLDATALOAD']

    def __init__(self):
        super().__init__()
        self.hashed_sload_cache = set()

    def _analyze(self, state: GlobalState, prev_state: Optional[GlobalState] = None) -> Optional[Result]:
        if prev_state and prev_state.instruction['opcode'] == 'CALLDATALOAD':
            state.mstate.stack[-1].annotate(CallData())
        elif prev_state and prev_state.instruction['opcode'] == 'SHA3' and self._has_annotation(state.mstate.stack[-1], CallData):
            state.mstate.stack[-1].annotate(HashedInput())
        elif prev_state and prev_state.instruction['opcode'] == 'SLOAD' and prev_state.instruction['address'] in self.hashed_sload_cache:
            self.hashed_sload_cache.remove(prev_state.instruction['address'])
            state.mstate.stack[-1].annotate(HashedStorage())

        if state.instruction['opcode'] == 'SLOAD' and self._has_annotation(state.mstate.stack[-1], HashedInput):
            # Workaround since annotations don't get propagated on loaded stack element.
            # Annotation gets added on SLOAD post_hook (see condition above).
            self.hashed_sload_cache.add(state.instruction['address'])
        elif state.instruction['opcode'] == 'JUMPI' and self._has_annotation(state.mstate.stack[-2], HashedStorage):
            return Result(state.environment.active_function_name)

        return None



import logging

from abc import ABC, abstractmethod
from typing import List, Optional, Set, Text, Type

from ithildin.report.analysis import ReportItem, Result

from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt.bitvec import BitVec

log = logging.getLogger(__name__)


class AnalysisStrategy(ABC):
    """
    Base class for contract analysis strategies. Subclasses can be instantiated by using the AnalysisStrategyFactory
    from the module \'ithildin.analysis.factory\'.

    When creating a new analysis strategy by subclassing this base class, override the *_analyze()* function.
    """

    pattern_name = ''
    report_title = ''
    report_description = ''

    pre_hooks: List[Text] = []
    post_hooks: List[Text] = []

    def __init__(self):
        self.cache: Set[Text] = set()
        self.results: List[ReportItem] = []

    def reset(self) -> None:
        self.cache = set()
        self.results = []

    def generate_report(self) -> ReportItem:
        report = ReportItem(self.report_title, self.report_description, self.pattern_name)
        for result in self.results:
            report.add_result(result)
        return report

    def execute(self, state: GlobalState) -> Optional[Result]:
        """ Execute analysis strategy on the given state. """
        if state.environment.active_function_name in self.cache:
            return None
        log.debug('Executing analysis strategy %s', type(self).__name__)
        result = self._analyze(state, state.node.states[-1] if len(state.node.states) > 0 else None)
        if result is not None:
            log.info('Analysis strategy %s got a hit in function %s', type(self).__name__, result.function_name)
            self.results.append(result)
            self.cache.add(state.environment.active_function_name)
        return result

    @abstractmethod
    def _analyze(self, state: GlobalState, prev_state: Optional[GlobalState] = None) -> Optional[Result]:
        """ Actual implementation of the analysis strategy. Override this when inheriting AnalysisStrategy. """
        pass

    def _has_annotation(self, bitvec: BitVec, annotation_type: Type) -> bool:
        """ Returns true if *bitvec* contains an annotation of type *annotation_type* """
        for annotation in bitvec.annotations:
            if isinstance(annotation, annotation_type):
                return True
        return False



import logging
import re

from ithildin.exception import ValidationError

from abc import ABC, abstractmethod
from typing import Optional, Text

from mythril.ethereum.evmcontract import EVMContract
from mythril.ethereum.interface.rpc.client import EthJsonRpc
from mythril.solidity.soliditycontract import SolidityContract
from mythril.support.loader import DynLoader

log = logging.getLogger(__name__)


class FileLoader(ABC):

    def __init__(self, file_path):
        self._file_path = file_path

    @abstractmethod
    def contract(self) -> EVMContract:
        pass


class BinaryLoader(FileLoader):

    def contract(self) -> EVMContract:
        try:
            with open(self._file_path) as contract_bin:
                bytecode = contract_bin.read()
        except IOError as e:
            log.error('Failed to open contract binary file: %s', e)
            raise IOError('Failed to open contract binary file')
        return EVMContract(creation_code=bytecode)


class SolidityLoader(FileLoader):

    def __init__(self, file_path, solc: Optional[Text] = 'solc'):
        super().__init__(file_path)
        self._solc = solc

    def contract(self) -> EVMContract:
        return SolidityContract(self._file_path, solc_binary=self._solc)


class JsonRpcLoader:

    def __init__(self, address: Text, rpc: Optional[Text] = None):
        assert address is not None, "No contract address provided"

        if rpc is None:
            eth_json_rpc = EthJsonRpc()
        else:
            match = re.match(r'(http(s)?:\/\/)?([a-zA-Z0-9\.\-]+)(:([0-9]+))?(\/.+)?', rpc)
            if match:
                host = match.group(3)
                port = match.group(5) if match.group(4) else None
                path = match.group(6) if match.group(6) else ''
                tls = bool(match.group(2))
                log.debug('Parsed RPC provider params: host=%s, port=%s, tls=%r, path=%s', host, port, tls, path)
                eth_json_rpc = EthJsonRpc(host=host + path, port=port, tls=tls)
            else:
                raise ValidationError('Invalid JSON RPC URL provided: "%s"' % rpc)
        self._dyn_loader = DynLoader(eth_json_rpc)
        self._address = address

    @property
    def dyn_loader(self) -> DynLoader:
        return self._dyn_loader

    @property
    def address(self) -> Text:
        return self._address



from enum import Enum
from typing import Optional
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt.bitvec import BitVec

from ithildin.analysis.base import AnalysisStrategy
from ithildin.report.analysis import Result


class Element(Enum):
    NUMBER = 1
    STORAGE = 2


class Comparison:
    """ Class to be used as annotation for comparison elements. """

    def __init__(self, item):
        self.item = item


class Storage:
    """ Class to be used for SLOAD elements. """

    def __init__(self, storage_address: Optional[int] = None):
        self.storage_address = storage_address


class BlockNumber:
    """ Class to be used as annotation for NUMBER elements. """

    def __init__(self, persist_to_world_state=False):
        self.persist_to_world_state = persist_to_world_state


class XConfirmation(AnalysisStrategy):

    pattern_name = 'X_CONFIRMATION'
    report_title = 'X-Confirmation (Block Count)'
    report_description = ('This pattern protects functions from being executed unless a condition related to the '
                          'current block number is met. For example one might want to wait for a certain number '
                          'of blocks to have passed before a transaction is marked as valid or confirmed.')

    pre_hooks = ['JUMPI', 'EQ', 'LT', 'GT']
    post_hooks = ['NUMBER', 'SLOAD']

    def _analyze(self, state: GlobalState, prev_state: Optional[GlobalState] = None) -> Optional[Result]:
        if prev_state and prev_state.instruction['opcode'] == 'NUMBER':
            state.mstate.stack[-1].annotate(BlockNumber())
        elif prev_state and prev_state.instruction['opcode'] == 'SLOAD':
            state.mstate.stack[-1].annotate(Storage(prev_state.mstate.stack[-1].value))

        if state.instruction['opcode'] in {'EQ', 'LT', 'GT'}:
            if self._has_annotation(state.mstate.stack[-1], BlockNumber) and self._has_annotation(state.mstate.stack[-2], Storage):
                state.mstate.stack[-1].annotate(Comparison(Element.NUMBER))
                state.mstate.stack[-2].annotate(Comparison(Element.STORAGE))
            elif self._has_annotation(state.mstate.stack[-1], Storage) and self._has_annotation(state.mstate.stack[-2], BlockNumber):
                state.mstate.stack[-1].annotate(Comparison(Element.STORAGE))
                state.mstate.stack[-2].annotate(Comparison(Element.NUMBER))
        elif state.instruction['opcode'] == 'JUMPI' and self._is_target_jumpi(state):
            storage_address = self._retrieve_storage_address(state.mstate.stack[-2])
            return Result(state.environment.active_function_name, storage_address=storage_address)

        return None

    def _is_target_jumpi(self, state: GlobalState) -> bool:
        """
        Helper method for checking if the second element on the stack before a JUMPI instruction
        contains the targeted annotations, i.e. two annotations of type Comparison with elements
        NUMBER and STORAGE.

        Returns
        -------
        True if the annotations and their elements are present, False otherwise.
        """
        item_flags = 0b00
        for annotation in state.mstate.stack[-2].annotations:
            if isinstance(annotation, Comparison):
                item_flags |= 0b01 if annotation.item == Element.NUMBER else 0
                item_flags |= 0b10 if annotation.item == Element.STORAGE else 0
        return item_flags == 0b11

    def _retrieve_storage_address(self, bitvec: BitVec) -> Optional[int]:
        """ Helper function to retrieve the *storage_address* attribute from a BitVec instance. """
        for annotation in bitvec.annotations:
            if isinstance(annotation, Storage):
                return annotation.storage_address
        return None



import logging
import time
from typing import Optional, Text, Union

from ithildin.analysis.loader import StrategyLoader
from ithildin.contract.loader import FileLoader, JsonRpcLoader
from ithildin.report.analysis import Report

from mythril.laser.ethereum import svm
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.support.loader import DynLoader


log = logging.getLogger(__name__)


class LaserWrapper:

    def __init__(self, strategy_loader: Optional[StrategyLoader] = StrategyLoader()):
        self.strategy_loader = strategy_loader

    def execute(self,
                timeout: Optional[float] = 60,
                max_depth: Optional[int] = 128,
                creation_code: Optional[Text] = None,
                target_address: Optional[Text] = None,
                dyn_loader: Optional[DynLoader] = None,
                contract_loader: Optional[Union[FileLoader, JsonRpcLoader]] = None) -> Report:
        if contract_loader is not None:
            if isinstance(contract_loader, FileLoader):
                creation_code = contract_loader.contract().creation_disassembly.bytecode
            elif isinstance(contract_loader, JsonRpcLoader):
                target_address = contract_loader.address
                dyn_loader = contract_loader.dyn_loader
            else:
                raise ValueError('Invalid type for contract_loader parameter')

        world_state = None
        if creation_code is not None and target_address is None:
            log.info('Running symbolic execution in creation mode...')
            laser = svm.LaserEVM(execution_timeout=timeout,
                                 max_depth=max_depth,
                                 requires_statespace=False)
        elif creation_code is None and target_address is not None:
            assert dyn_loader is not None, "Dynamic Loader has not been provided"
            log.info('Running symbolic execution in existing mode...')
            laser = svm.LaserEVM(dynamic_loader=dyn_loader,
                                 execution_timeout=timeout,
                                 max_depth=max_depth,
                                 requires_statespace=False)
            world_state = WorldState()
            world_state.accounts_exist_or_load(target_address, dyn_loader)
        else:
            raise ValueError('Either creation_code or target_address needs to be provided')

        for strategy in self.strategy_loader.get_strategies():
            for hook in strategy.pre_hooks:
                laser.register_hooks('pre', {hook: [strategy.execute]})
            for hook in strategy.post_hooks:
                laser.register_hooks('post', {hook: [strategy.execute]})

        # Run symbolic execution
        start_time = time.time()
        laser.sym_exec(creation_code=creation_code,
                       contract_name='Unknown',
                       world_state=world_state,
                       target_address=int(target_address, 16) if target_address else None)
        log.info('Symbolic execution finished in %.2f seconds.', time.time() - start_time)

        report = Report(start_time=start_time, end_time=time.time())
        report.contract_code = creation_code
        report.contract_address = target_address
        for strategy in self.strategy_loader.get_strategies():
            report.add_report(strategy.generate_report())
        self._post_process_report(report, target_address, dyn_loader)
        return report

    def _post_process_report(self, report: Report, target_address: Text, dyn_loader: DynLoader) -> None:
        if dyn_loader is None or target_address is None:
            return
        for report_item in report.reports:
            for result in report_item.results:
                if result.storage_address is not None:
                    result.storage_content = dyn_loader.read_storage(target_address, result.storage_address)



import logging

from argparse import ArgumentParser
from typing import Text

from ithildin import __version__
from ithildin.analysis.symbolic import LaserWrapper
from ithildin.contract.loader_factory import get_factory, LoaderFactoryType
from ithildin.support.compiler_version import VersionParseAction
from ithildin.tools.benchmark import benchmark, STRATEGIES

# Default analysis arguments
DEFAULT_MAX_DEPTH = 128
DEFAULT_RPC = 'http://127.0.0.1:8545'
DEFAULT_SOLC = 'solc'
DEFAULT_TIMEOUT_ANALYSIS = 60

# Default benchmark arguments
DEFAULT_DELIMITER = ','
DEFAULT_HAS_HEADER = True
DEFAULT_ADDRESS_COLUMN = 0
DEFAULT_SAMPLE_SIZE = 5
DEFAULT_TIMEOUT_BENCHMARK = 90
DEFAULT_SEED = 1
DEFAULT_VERIFICATION_RATIO = 0.1


def populate_analysis_parser(parser: ArgumentParser) -> None:
    parser.add_argument('--json', action='store_true', dest='as_json', help='print report as JSON to standard output')

    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument('-a', '--address', metavar='ADDRESS', type=Text, help='contract address to analyze')
    input_group.add_argument('-s', '--sol', metavar='PATH', type=Text, dest='sol_path', help='path to solidity contract')
    input_group.add_argument('-b', '--bin', metavar='PATH', type=Text, dest='bin_path',
                             help='path to file containing contract creation bytecode')

    sym_exec_arguments = parser.add_argument_group('symbolic execution arguments')
    sym_exec_arguments.add_argument('--timeout', metavar='SEC', type=int, default=DEFAULT_TIMEOUT_ANALYSIS,
                                    help='symbolic execution timeout (default: {})'.format(DEFAULT_TIMEOUT_ANALYSIS))
    sym_exec_arguments.add_argument('--max-depth', metavar='DEPTH', type=int, default=DEFAULT_MAX_DEPTH,
                                    help='max graph depth (default: {})'.format(DEFAULT_MAX_DEPTH))

    networking_group = parser.add_argument_group('networking arguments')
    networking_group.add_argument('--rpc', metavar="RPC", type=Text, default=DEFAULT_RPC,
                                  help='JSON RPC provider URL (default: \'{}\')'.format(DEFAULT_RPC))

    compilation_group = parser.add_argument_group('compilation arguments')
    compilation_group.add_argument('--solc', metavar='SOLC', type=Text, default=DEFAULT_SOLC,
                                   help='solc binary path (default: \'{}\')'.format(DEFAULT_SOLC))


def populate_benchmark_parser(parser: ArgumentParser) -> None:
    parser.add_argument('filename', metavar='FILE', type=str, help='the csv file containing contract instances')
    parser.add_argument('--infura-project', dest='infura_project_id', metavar='PROJECT_ID', type=str, required=True,
                        help='the Infura project ID for retrieving contract data from the mainchain')
    parser.add_argument('--strategy', default=list(STRATEGIES.keys())[0], choices=STRATEGIES.keys(),
                        help='which strategy to benchmark (default: \'{}\')'.format(list(STRATEGIES.keys())[0]))
    parser.add_argument('--timeout', metavar='SEC', type=int, default=DEFAULT_TIMEOUT_BENCHMARK,
                        help='the execution timeout for each contract (default: {})'.format(DEFAULT_TIMEOUT_BENCHMARK))
    parser.add_argument('--max-depth', metavar='DEPTH', type=int, default=DEFAULT_MAX_DEPTH,
                        help='max graph depth (default: {})'.format(DEFAULT_MAX_DEPTH))

    sampling_group = parser.add_argument_group('sampling options')
    sampling_group.add_argument('--sample-size', metavar='SIZE', type=int, default=DEFAULT_SAMPLE_SIZE,
                                help='the sample size to be picked from the CSV instances (default: {})'.format(DEFAULT_SAMPLE_SIZE))
    sampling_group.add_argument('--random-seed', metavar='SEED', type=int, default=DEFAULT_SEED,
                                help='a seed for the sampling RNG (default: {})'.format(DEFAULT_SEED))
    sampling_group.add_argument('--verification-ratio', metavar='RATIO', type=float, default=DEFAULT_VERIFICATION_RATIO,
                                help='the ratio of the sampled contracts to manually verify (default: {})'.format(DEFAULT_VERIFICATION_RATIO))
    sampling_group.add_argument('--compiler-target', metavar='x.x.x', action=VersionParseAction,
                                help='the Solidity compiler target to match in the sample (e.g. ^0.7.0)')

    csv_group = parser.add_argument_group('CSV arguments')
    csv_group.add_argument('--has-header', action='store_true', default=DEFAULT_HAS_HEADER,
                           help='does the CSV file contain a header (default: {})'.format(DEFAULT_HAS_HEADER))
    csv_group.add_argument('--csv-delimiter', metavar='DELIMITER', type=str, default=DEFAULT_DELIMITER,
                           help='the CSV delimiter (default: \'{}\')'.format(DEFAULT_DELIMITER))
    csv_group.add_argument('--address-column', metavar='COL', type=int, default=DEFAULT_ADDRESS_COLUMN,
                           help='column index that contains the contract addresses (default: {})'.format(DEFAULT_ADDRESS_COLUMN))
    csv_group.add_argument('--compiler-column', metavar='COL', type=int, help='the column containing the compiler name')
    csv_group.add_argument('--version-column', metavar='COL', type=int,
                           help='column index that contains the compiler version used')


def get_parser() -> ArgumentParser:
    program_name = 'Ithildin - EVM bytecode semantic analysis tool based on Mythril'
    parser = ArgumentParser(description=program_name)
    parser.add_argument('--version', action='version', version='Ithildin v{}'.format(__version__))
    parser.add_argument('--verbose', action='store_true', help='print debugging output')
    subparsers = parser.add_subparsers(dest='command', help='Commands')
    # Add analysis parser
    analysis_parser = subparsers.add_parser('analyze', help='begin analysis of a contract')
    populate_analysis_parser(analysis_parser)
    # Add benchmark parser
    benchmark_parser = subparsers.add_parser('benchmark', help='execute benchmarking tool')
    populate_benchmark_parser(benchmark_parser)

    return parser


def analyze(args) -> None:
    # Get the contract loader factory based on the specified options
    if args.bin_path:
        contract_loader_factory = get_factory(LoaderFactoryType.BINARY, path=args.bin_path)
    elif args.sol_path:
        contract_loader_factory = get_factory(LoaderFactoryType.SOLIDITY, path=args.sol_path, solc=args.solc)
    elif args.address:
        contract_loader_factory = get_factory(LoaderFactoryType.JSON_RPC, address=args.address, rpc=args.rpc)
    else:
        raise NotImplementedError('This feature hasn\'t been implemented yet')

    contract_loader = contract_loader_factory.create()
    symbolic_analysis = LaserWrapper()
    report = symbolic_analysis.execute(contract_loader=contract_loader, timeout=args.timeout, max_depth=args.max_depth)
    print(report.to_json(pretty=True) if args.as_json else report.to_text())


def main():
    parser = get_parser()
    args = parser.parse_args()
    # Set logging level to DEBUG for all logers if the *--verbose* option was specified
    if args.verbose:
        for logger in [logging.getLogger(name) for name in logging.root.manager.loggerDict]:
            if args.verbose:
                logger.setLevel(logging.DEBUG)

    if args.command == 'analyze':
        analyze(args)
    elif args.command == 'benchmark':
        benchmark(args)



from typing import List

from .base import AnalysisStrategy
from .strategies.ownership import Ownership
from .strategies.x_confirmation import XConfirmation

from ithildin.support.singleton import Singleton


class StrategyLoader(metaclass=Singleton):

    def __init__(self) -> None:
        self.strategies = []
        self.strategies.extend(self.default_strategies())

    def get_strategies(self) -> List[AnalysisStrategy]:
        return self.strategies

    def set_strategies(self, strategies: List[AnalysisStrategy]) -> None:
        assert strategies is not None and len(strategies) > 0
        self.strategies = strategies

    def register_strategy(self, strategy: AnalysisStrategy) -> None:
        assert strategy is not None, 'No strategy provided'
        self.strategies.append(strategy)

    def reset_strategies(self) -> None:
        for strategy in self.strategies:
            strategy.reset()

    def default_strategies(self) -> List[AnalysisStrategy]:
        return [
            Ownership(),
            XConfirmation()
        ]



import json

from jinja2 import Environment, PackageLoader
from typing import List, Optional, Text

from ithildin import __version__


class Result:

    def __init__(self,
                 contract_address: Text,
                 contract_index: int,
                 detected_functions: List[Text],
                 compiler_version: Optional[Text] = None) -> None:
        self.contract_address = contract_address
        self.contract_index = contract_index
        self.detected_functions = detected_functions
        self.compiler_version = compiler_version
        self.verified = False
        self.true_positives = 0
        self.false_positives = 0
        self.true_negatives = 0
        self.false_negatives = 0

    @property
    def total_hits(self) -> int:
        return len(self.detected_functions)

    @property
    def total_functions_count(self):
        return self.true_positives + self.false_positives + self.true_negatives + self.false_negatives

    def to_dict(self):
        return {
            'contractAddress': self.contract_address,
            'contractIndex': self.contract_index,
            'detectedFunctions': self.detected_functions
        }

    def to_json(self, pretty=False):
        return json.dumps(self.to_dict(), indent=2 if pretty else None)


class Report:

    def __init__(self, strategy_name: Text, random_seed: int, exec_timeout: int, max_depth: int,
                 verification_ratio: float, target_version=None, contracts_filename=None,
                 file_sha256sum=None, start_time=None, end_time=None) -> None:
        self.strategy_name = strategy_name
        self.random_seed = random_seed
        self.exec_timeout = exec_timeout
        self.max_depth = max_depth
        self.verification_ratio = verification_ratio
        self.target_version = target_version
        self.contracts_filename = contracts_filename
        self.file_sha256sum = file_sha256sum
        self.start_time = start_time
        self.end_time = end_time
        self._results: List[Result] = []

    @property
    def results(self) -> List[Result]:
        return self._results

    def add_result(self, result: Result) -> None:
        assert result is not None
        self._results.append(result)

    @property
    def sample_size(self) -> int:
        return len(self.results)

    @property
    def total_detections(self) -> int:
        return sum(result.total_hits for result in self.results)

    @property
    def true_positives(self) -> int:
        return sum(result.true_positives for result in self.results if result.verified)

    @property
    def false_positives(self) -> int:
        return sum(result.false_positives for result in self.results if result.verified)

    @property
    def true_negatives(self) -> int:
        return sum(result.true_negatives for result in self.results if result.verified)

    @property
    def false_negatives(self) -> int:
        return sum(result.false_negatives for result in self.results if result.verified)

    @property
    def precision(self) -> Optional[float]:
        try:
            return self.true_positives / (self.true_positives + self.false_positives)
        except ZeroDivisionError:
            return None

    @property
    def recall(self) -> Optional[float]:
        try:
            return self.true_positives / (self.true_positives + self.false_negatives)
        except ZeroDivisionError:
            return None

    def to_markdown(self):
        environment = Environment(loader=PackageLoader('ithildin.report'), trim_blocks=True)
        template = environment.get_template('benchmark_report.md.jinja2')
        return template.render(report=self, program_version=__version__)

    def __repr__(self) -> Text:
        return (
            '<Report '
            'strategy_name={0.strategy_name} '
            'random_seed={0.random_seed} '
            'exec_timeout={0.exec_timeout} '
            'max_depth={0.max_depth} '
            'verification_ratio={0.verification_ratio} '
            'target_version={0.target_version} '
            'contracts_filename={0.contracts_filename} '
            'file_sha256sum={0.file_sha256sum} '
            'start_time={0.start_time} '
            'end_time={0.end_time} '
            'results={0.results}'
            '>'
        ).format(self)



import logging

from mythril.mythril.mythril_config import MythrilConfig

__version__ = '0.1.4'

formatter = logging.Formatter('[%(levelname)s\t] %(asctime)s - %(name)s %(message)s')

stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.DEBUG)
stream_handler.setFormatter(formatter)

log = logging.getLogger(__name__)
log.addHandler(stream_handler)
log.setLevel(logging.INFO)

# Ensure the mythril home directory '~/.mythril' is present and populated.
# This is an issue that arises when mythril is used as a library and tries to access ~/.mythril/signatures.db
# Related issue: https://github.com/metagon/ithildin/issues/1
MythrilConfig()



import json
from jinja2 import Environment, PackageLoader
from typing import Dict, List, Optional, Text


class Result:

    def __init__(self, function_name: Text, storage_address: Optional[int] = None, storage_content: Optional[Text] = None) -> None:
        self.function_name = function_name
        self.storage_address = storage_address
        self.storage_content = storage_content

    def to_dict(self) -> Dict:
        as_dict = {'functionName': self.function_name}
        if self.storage_address is not None:
            as_dict['storageAddress'] = self.storage_address
        if self.storage_content is not None:
            as_dict['storageContent'] = self.storage_content
        return as_dict

    def __repr__(self):
        return (
            '<Result '
            'function_name={0.function_name} '
            'storage_address={0.storage_address} '
            'storage_content={0.storage_content}'
            '>'
        ).format(self)


class ReportItem:

    def __init__(self, title: Text, description: Text, pattern_name: Text) -> None:
        self.title = title
        self.description = description
        self.pattern_name = pattern_name
        self.results: List[Result] = []

    def add_result(self, result: Result) -> None:
        self.results.append(result)

    def to_dict(self) -> Dict:
        return {
            'patternName': self.pattern_name,
            'title': self.title,
            'description': self.description,
            'results': [result.to_dict() for result in self.results]
        }

    def __repr__(self):
        return (
            '<ReportItem '
            'title={0.title} '
            'description={0.description} '
            'pattern_name={0.pattern_name} '
            'results={0.results}'
            '>'
        ).format(self)


class Report:

    def __init__(self, start_time: Optional[float] = None, end_time: Optional[float] = None) -> None:
        self.start_time = start_time
        self.end_time = end_time
        self.contract_address = None
        self.contract_code = None
        self.reports = []

    def add_report(self, report: ReportItem) -> None:
        if report is not None:
            self.reports.append(report)

    def add_all(self, items: List[ReportItem]) -> None:
        self.reports.extend(items)

    def to_dict(self) -> Dict:
        as_dict = {
            'startTime': self.start_time,
            'endTime': self.end_time
        }
        if self.contract_address is not None:
            as_dict['contractAddress'] = self.contract_address
        if self.contract_code is not None:
            as_dict['contractCode'] = self.contract_code
        as_dict['reports'] = [report.to_dict() for report in self.reports if len(report.results) > 0]
        return as_dict

    def to_text(self) -> Text:
        environment = Environment(loader=PackageLoader('ithildin.report'), trim_blocks=True)
        template = environment.get_template('analysis_report.txt.jinja2')
        return template.render(report=self)

    def to_json(self, pretty: bool = False) -> Text:
        return json.dumps(self.to_dict(), indent=2 if pretty else None)

    def __repr__(self):
        return (
            '<Report '
            'start_time={0.start_time} '
            'end_time={0.end_time} '
            'reports={0.reports}'
            '>'
        ).format(self)



from enum import Enum
from typing import Optional
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt.bitvec import BitVec

from ithildin.analysis.base import AnalysisStrategy
from ithildin.report.analysis import Result

import logging
log = logging.getLogger(__name__)


class Actor(Enum):
    SENDER = 1
    OWNER = 2


class Caller:
    """ Class to be used as annotation for CALLER elements. """
    ...


class Storage:
    """ Class to be used as annotation for SLOAD elements. """

    def __init__(self, storage_address: Optional[int] = None):
        self.storage_address = storage_address


class Equality:
    """ Class to be used as annotation for EQ elements. """

    def __init__(self, actor: Actor):
        self.actor = actor


class Ownership(AnalysisStrategy):

    pattern_name = 'OWNERSHIP'
    report_title = 'Ownership'
    report_description = ('This pattern aims at restricting functions to specific addresses. For example '
                          'the owner of the contract is specified in the constructor and only that account '
                          'is allowed to access a function and change the contract\'s state.')

    pre_hooks = ['EQ', 'JUMPI']
    post_hooks = ['CALLER', 'SLOAD']

    def _analyze(self, state: GlobalState, prev_state: Optional[GlobalState] = None) -> Optional[Result]:
        if prev_state and prev_state.instruction['opcode'] == 'CALLER':
            state.mstate.stack[-1].annotate(Caller())
        elif prev_state and prev_state.instruction['opcode'] == 'SLOAD':
            state.mstate.stack[-1].annotate(Storage(prev_state.mstate.stack[-1].value))

        if state.instruction['opcode'] == 'EQ':
            # Check if both top stack elemnts have been annotated with Caller and Storage,
            # in which case we annotate both elements with the Equality annotation, and
            # their respective actors.
            if self._has_annotation(state.mstate.stack[-1], Storage) and self._has_annotation(state.mstate.stack[-2], Caller):
                state.mstate.stack[-1].annotate(Equality(Actor.OWNER))
                state.mstate.stack[-2].annotate(Equality(Actor.SENDER))
            elif self._has_annotation(state.mstate.stack[-1], Caller) and self._has_annotation(state.mstate.stack[-2], Storage):
                state.mstate.stack[-1].annotate(Equality(Actor.SENDER))
                state.mstate.stack[-2].annotate(Equality(Actor.OWNER))
        elif state.instruction['opcode'] == 'JUMPI' and self._is_target_jumpi(state):
            storage_address = self._retrieve_storage_address(state.mstate.stack[-2])
            return Result(state.environment.active_function_name, storage_address=storage_address)

        return None

    def _is_target_jumpi(self, state: GlobalState) -> bool:
        """
        Helper method for checking if the JUMPI contains the targeted annotations.

        Returns
        -------
        True if the second stack element contains two annotations of type Equality with attributes
        'owner' and 'sender', False otherwise.
        """
        actor_flags = 0b00
        for annotation in state.mstate.stack[-2].annotations:
            if isinstance(annotation, Equality):
                actor_flags |= 0b01 if annotation.actor == Actor.OWNER else 0
                actor_flags |= 0b10 if annotation.actor == Actor.SENDER else 0
        return actor_flags == 0b11

    def _retrieve_storage_address(self, bitvec: BitVec) -> Optional[int]:
        """ Helper function to retrieve the *storage_address* attribute from a BitVec instance. """
        for annotation in bitvec.annotations:
            if isinstance(annotation, Storage):
                return annotation.storage_address
        return None



import pathlib
from setuptools import find_packages, setup

VERSION = '0.1.4'
CURRENT_DIR = pathlib.Path(__file__).parent
README = (CURRENT_DIR / 'README.md').read_text()
PYTHON_REQUIREMENT = '>=3.6.0'

REQUIREMENTS = [
    "Jinja2==2.11.2",
    "mythril==0.22.14"
]

setup(
    name='ithildin',
    version=VERSION,
    description='Semantic analyzer of EVM bytecode based on Mythril',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/metagon/ithildin',
    author='Philippos Gorgoris',
    author_email='philippos@gorgoris.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    packages=find_packages(exclude=('scripts')),
    include_package_data=True,
    install_requires=REQUIREMENTS,
    python_requires=PYTHON_REQUIREMENT,
    entry_points={'console_scripts': ['ithil=ithildin.__main__:main']}
)



import csv
import hashlib
import logging
import math
import os
import random
import time

from functools import lru_cache
from typing import Optional, Set, Text

from ithildin.analysis.loader import StrategyLoader
from ithildin.analysis.symbolic import LaserWrapper
from ithildin.analysis.loader import Ownership, XConfirmation
from ithildin.contract.loader_factory import get_factory, LoaderFactoryType
from ithildin.report.benchmark import Report, Result
from ithildin.support.compiler_version import Version, VersionMatcher

log = logging.getLogger(__name__)

TIME_FORMAT = '%Y-%m-%d %H:%M:%S (%z)'
STRATEGIES = {
    'ownership': Ownership(),
    'x-confirmation': XConfirmation()
}


def start_verification(report: Report, verification_sample: Set[int]) -> None:
    print()
    print('=' * 80)
    print('! Entering verification mode...')
    for result in report.results:
        if result.contract_index not in verification_sample:
            continue
        print('! Verifying contract at address %s' % result.contract_address)
        print('! Results:\n' + result.to_json(pretty=True))
        # TODO: Does mythril contain any function for retrieving this value?
        print('! How many functions does the contract contain in total (excluding constructors)?')
        functions_count_answer = input('> Your answer [Type \'s\' to skip]: ')
        if functions_count_answer == 's':
            continue
        total_functions_count = int(functions_count_answer, base=10)
        if result.total_hits > 0:
            print('! Did the analysis strategy correctly identify *all* functions in the contract?')
            answer = input('> Your answer [y/n]: ')
            while answer not in {'y', 'n'}:
                answer = input('> Enter a valid answer [y/n]: ')
            if answer == 'y':
                result.true_positives = result.total_hits
                result.true_negatives = total_functions_count - result.total_hits
            else:
                false_positive_count = int(input('> How many FALSE POSITIVES: '), base=10)
                false_negative_count = int(input('> How many FALSE NEGATIVES: '), base=10)
                result.true_positives = result.total_hits - false_positive_count
                result.false_positives = false_positive_count
                result.true_negatives = total_functions_count - false_negative_count - result.total_hits
                result.false_negatives = false_negative_count
        else:
            print('! The analysis strategy reported no results for this contract, is that correct?')
            answer = input('> Your answer [y/n]: ')
            while answer not in {'y', 'n'}:
                answer = input('> Enter a valid answer [y/n]: ')
            if answer == 'y':
                result.true_negatives = total_functions_count
            else:
                false_negative_count = int(input('> How many FALSE NEGATIVES: '), base=10)
                result.true_negatives = total_functions_count - false_negative_count
                result.false_negatives = false_negative_count
        result.verified = True
    print('=' * 80)
    print(report.to_markdown())


@lru_cache(maxsize=5)
def count_rows(file: Text, delimiter=',') -> int:
    with open(file, 'r') as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=delimiter)
        return sum(1 for _ in csv_reader)


def generate_contract_sample(file: Text,
                             sample_size: int,
                             address_column: int,
                             compiler_name_column: Optional[int] = None,
                             compiler_version_column: Optional[int] = None,
                             has_header: Optional[bool] = True,
                             delimiter: Optional[Text] = ',',
                             version_matcher: Optional[VersionMatcher] = None) -> Set[int]:
    if version_matcher is not None:
        assert compiler_version_column is not None, 'Compiler version column not provided'
        matched_version_indexes = set()
        with open(file, 'r') as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=delimiter)
            for i, row in enumerate(csv_reader):
                if compiler_name_column is None or row[compiler_name_column].strip() == 'Solidity':
                    try:
                        version = Version(raw=row[compiler_version_column])
                    except ValueError:
                        continue
                    if version_matcher.matches(version):
                        matched_version_indexes.add(i)
        return set(random.sample(matched_version_indexes, sample_size))
    else:
        row_count = count_rows(file, delimiter=delimiter)
        return set(random.sample(range(1 if has_header else 0, row_count), sample_size))


def benchmark(args) -> None:
    random.seed(args.random_seed)
    instance_count = count_rows(args.filename, delimiter=args.csv_delimiter) - 1 if args.has_header else 0
    file_sha256sum = hashlib.sha256(open(args.filename, 'rb').read()).hexdigest()
    contract_sample = generate_contract_sample(args.filename, args.sample_size, args.address_column,
                                               compiler_name_column=args.compiler_column, compiler_version_column=args.version_column,
                                               has_header=args.has_header, delimiter=args.csv_delimiter, version_matcher=args.compiler_target)
    benchmark_report = Report(args.strategy.capitalize(), args.random_seed, args.timeout, args.max_depth, args.verification_ratio,
                              contracts_filename=os.path.basename(args.filename), file_sha256sum=file_sha256sum,
                              start_time=time.strftime(TIME_FORMAT), target_version=args.compiler_target.raw if args.compiler_target else None)
    rpc = 'https://mainnet.infura.io/v3/' + args.infura_project_id
    strategy_loader = StrategyLoader()
    strategy_loader.set_strategies([STRATEGIES[args.strategy]])
    positive_instances = set()
    with open(args.filename, 'r') as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=args.csv_delimiter)
        for i, row in enumerate(csv_reader):
            if i not in contract_sample:
                continue
            target_address = row[args.address_column]
            log.info('Analyzing contract %d/%d at address %s', i + 1, instance_count, target_address)
            loader_factory = get_factory(LoaderFactoryType.JSON_RPC, address=target_address, rpc=rpc)
            analysis_report = LaserWrapper().execute(contract_loader=loader_factory.create(), timeout=args.timeout, max_depth=args.max_depth)
            if sum(len(report_item.results) for report_item in analysis_report.reports) > 0:
                positive_instances.add(i)
            else:
                log.info('Nothing found for contract %d/%d at address %s', i + 1, instance_count, target_address)
            function_names = [result.function_name
                              for report_item in analysis_report.reports if len(report_item.results) > 0
                              for result in report_item.results]
            compiler_version = row[args.version_column] if args.version_column is not None else None
            benchmark_report.add_result(Result(target_address, i, function_names, compiler_version=compiler_version))
            strategy_loader.reset_strategies()
    benchmark_report.end_time = time.strftime(TIME_FORMAT)
    negative_instances = contract_sample - positive_instances
    positive_sample = set(random.sample(positive_instances, math.ceil(len(positive_instances) * args.verification_ratio)))
    negative_sample = set(random.sample(negative_instances, math.ceil(len(negative_instances) * args.verification_ratio)))
    start_verification(benchmark_report, positive_sample | negative_sample)



import re

from argparse import Action
from typing import Optional, Text

VERSION_REXEG = r'v?(\d)\.(\d{1,2})\.(\d{1,2})'
VERSION_MATCHER_REGEX = r'(\^|>|>=|==)?\s*(\d)\.(\d{1,2})\.(\d{1,2})(\,\s*(<|<=)\s*(\d)\.(\d{1,2})\.(\d{1,2}))?'

COMPARE_PREDICATES = {
    '^': lambda x, y: x.major == y.major and x.minor == y.minor and x.hotfix >= y.hotfix,
    '>': lambda x, y: ((x.major > y.major) or
                       (x.major == y.major and x.minor > y.minor) or
                       (x.major == y.major and x.minor == y.minor and x.hotfix > y.hotfix)),
    '<': lambda x, y: ((x.major < y.major) or
                       (x.major == y.major and x.minor < y.minor) or
                       (x.major == y.major and x.minor == y.minor and x.hotfix < y.hotfix)),
    '>=': lambda x, y: ((x.major > y.major) or
                        (x.major == y.major and x.minor > y.minor) or
                        (x.major == y.major and x.minor == y.minor and x.hotfix >= y.hotfix)),
    '<=': lambda x, y: ((x.major < y.major) or
                        (x.major == y.major and x.minor < y.minor) or
                        (x.major == y.major and x.minor == y.minor and x.hotfix <= y.hotfix)),
    '==': lambda x, y: x.major == y.major and x.minor == y.minor and x.hotfix == y.hotfix
}


class Version:

    def __init__(self,
                 major: Optional[int] = None,
                 minor: Optional[int] = None,
                 hotfix: Optional[int] = None,
                 raw: Optional[Text] = None):
        if major is not None and minor is not None and hotfix is not None:
            self.raw = '{}.{}.{}'.format(major, minor, hotfix)
            self.major = major
            self.minor = minor
            self.hotfix = hotfix
        elif raw is not None:
            self.raw = raw
            self._parse_raw(raw)
        else:
            raise ValueError('Either (major, minor and hotfix) values, or raw string must be given')

    def _parse_raw(self, raw: Text) -> None:
        match = re.match(VERSION_REXEG, raw)
        if match:
            self.major = int(match.group(1), base=10)
            self.minor = int(match.group(2), base=10)
            self.hotfix = int(match.group(3), base=10)
        else:
            raise ValueError('Invalid version pattern: %s' % raw)

    def __repr__(self):
        return (
            '<Version '
            'major={0.major} '
            'minor={0.minor} '
            'hotfix={0.hotfix}'
            '>'
        ).format(self)


class VersionMatcher:

    def __init__(self, raw: Text) -> None:
        self.raw = raw
        self.compare_lo = None
        self.compare_hi = None
        self.version_lo = None
        self.version_hi = None
        self._parse_raw(raw)

    def matches(self, version: Version) -> bool:
        matches_lo = self.compare_lo(version, self.version_lo)
        matches_hi = True
        if self.version_hi is not None:
            matches_hi = self.compare_hi(version, self.version_hi)
        return matches_lo and matches_hi

    def _parse_raw(self, raw: Text) -> None:
        match = re.match(VERSION_MATCHER_REGEX, raw)
        if match:
            self.compare_lo = COMPARE_PREDICATES[match.group(1)] if match.group(1) else COMPARE_PREDICATES['==']
            self.version_lo = Version(int(match.group(2), base=10), int(match.group(3), base=10), int(match.group(4), base=10))
            if match.group(5):
                self.compare_hi = COMPARE_PREDICATES[match.group(6)]
                self.version_hi = Version(int(match.group(7), base=10), int(match.group(8), base=10), int(match.group(9), base=10))
        else:
            raise ValueError('Invalid version match pattern: %s' % raw)


class VersionParseAction(Action):

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError('nargs not allowed here')
        super(VersionParseAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, VersionMatcher(values))
